#include <cstdint>
#include <algorithm>

#include <QString>
#include <QDebug>

#include "tslib.h"
#include "sendWRcmd.h"
#include "shared_mem_buffer.h"

void indat_PrnPwr(void);

/* convention: use simple (not rotating) FIFO Stack:
Example: nrOfCmdsInQueue=4 then
    nextAsynchsendCmd0[0]=cmd1      // was stored as first
    nextAsynchsendCmd0[1]=cmd2
    nextAsynchsendCmd0[2]=cmd3
    nextAsynchsendCmd0[3]=cmd4      // came in as last

    Send: [0] first, then move buffer 1 down:
    nextAsynchsendCmd0[0]=cmd2
    nextAsynchsendCmd0[1]=cmd3
    nextAsynchsendCmd0[2]=cmd4
    nextAsynchsendCmd0[3]=0;
    nrOfCmdsInQueue=3 now
*/
template <typename T, std::size_t N>
static void *move_buffer_down_by_one(T (&buffer)[N]) {
    return memmove(buffer, &buffer[1], sizeof(buffer)-sizeof(buffer[0]));
}
template <typename T, std::size_t N>
static void clear(T (&buffer)[N]) {
    memset((char *)(&buffer[0]), 0x00, sizeof(buffer));
}
template <typename T, std::size_t M, std::size_t N>
static void clear(T (&buffer)[M][N]) {
    memset((char *)(&buffer[0][0]), 0x00, sizeof(buffer));
}
template <typename T>
static void sendWRcmd_clearCmdStack(T &cmdStack) {
    clear(cmdStack.AsynchSend);
    clear(cmdStack.para1);
    clear(cmdStack.para2);
    clear(cmdStack.para3);
    clear(cmdStack.para4);
    cmdStack.nrOfCmdsInQueue = 0;
}
template <typename T, typename T1, typename T2, typename T3, typename T4>
static bool sendWRcmd_setSendCmd(T &cmdStack, uint16_t nextCmd, T1 dat1,
                                 T2 dat2, T3 dat3, T4 dat4) {
    uint8_t const &n = cmdStack.nrOfCmdsInQueue;
    int const STACKDEPTH = sizeof(cmdStack.para1);
    if (n < STACKDEPTH) {
        cmdStack.AsynchSend[n] = nextCmd;
        cmdStack.para1[n] = dat1;
        cmdStack.para2[n] = dat2;
        cmdStack.para3[n] = dat3;
        cmdStack.para4[n] = dat4;
        cmdStack.nrOfCmdsInQueue += 1;
        return true;                // ok, will be sent
    }
    qCritical() << "cannot save cmd because stack is full";
    return false;           // not possible
}
template <typename T, typename T1, typename T2, typename T3, typename T4>
static bool sendWRcmd_getSendCmd(T &cmdStack, T1 *dat1, T2 *dat2,
                                 T3 *dat3, T4 *dat4) {
    if (dat1 && dat2 && dat3 && dat4) {
        uint8_t const &n = cmdStack.nrOfCmdsInQueue;
        int const STACKDEPTH = sizeof(cmdStack.para1);
        if ((n > 0) && (n <= STACKDEPTH)) {
            uint16_t const nxt = cmdStack.AsynchSend[0];
            move_buffer_down_by_one(cmdStack.AsynchSend);
            *dat1 = cmdStack.para1[n];
            *dat2 = cmdStack.para2[n];
            *dat3 = cmdStack.para3[n];
            *dat4 = cmdStack.para4[n];
            cmdStack.nrOfCmdsInQueue = n - 1;
            return nxt;
        }
    }
    qCritical() << "cannot fetch cmd";
    return 0;   // error
}

void sendWRcmd_INI(void) {
    sendWRcmd_clrCmdStack();
    sendWRcmd_clrCmd4Stack();
    sendFDcmd_clrStack();
    longFDcmd_clrStack();
}

void sendWRcmd_clrCmdStack(void) {
    sendWRcmd_clearCmdStack(SharedMemBuffer::getData()->Cmd0);
}
void sendWRcmd_clrCmd4Stack(void) {
    sendWRcmd_clearCmdStack(SharedMemBuffer::getData()->Cmd4);
}
void sendWRcmd_clrCmd8Stack(void) {
    sendWRcmd_clearCmdStack(SharedMemBuffer::getData()->Cmd8);
}

bool sendWRcmd_setSendCommand0(uint16_t nextCmd, uint8_t dat1, uint8_t dat2,
                               uint8_t dat3, uint8_t dat4) {
    // write Command to memory, wait for transport
    return sendWRcmd_setSendCmd(SharedMemBuffer::getData()->Cmd0,
                                nextCmd, dat1, dat2, dat3, dat4);
}
bool sendWRcmd_setSendCommand4(uint16_t nextCmd,
                               uint8_t dat1, uint8_t dat2,
                               uint8_t dat3, uint8_t dat4) {
    // write Command to memory, wait for transport
    return sendWRcmd_setSendCmd(SharedMemBuffer::getData()->Cmd4, nextCmd, dat1,
                                dat2, dat3, dat4);
}
bool sendWRcmd_setSendCommand8(uint16_t nextCmd,
                               uint8_t dat1, uint8_t dat2,
                               uint16_t dat3, uint32_t dat4) {
    // write Command to memory, wait for transport
    return sendWRcmd_setSendCmd(SharedMemBuffer::getData()->Cmd8, nextCmd, dat1,
                                dat2, dat3, dat4);
}

uint16_t sendWRcmd_getSendCommand0(uint8_t *dat1, uint8_t *dat2,
                                   uint8_t *dat3, uint8_t *dat4) {
    return sendWRcmd_getSendCmd(SharedMemBuffer::getData()->Cmd0, dat1, dat2,
                                dat3, dat4);
}
uint16_t sendWRcmd_getSendCommand4(uint8_t *dat1, uint8_t *dat2,
                                   uint8_t *dat3, uint8_t *dat4) {
    return sendWRcmd_getSendCmd(SharedMemBuffer::getData()->Cmd4, dat1, dat2,
                                dat3, dat4);
}
uint16_t sendWRcmd_getSendCommand8(uint8_t *dat1, uint8_t *dat2,
                                   uint16_t *dat3, uint32_t *dat4) {
    return sendWRcmd_getSendCmd(SharedMemBuffer::getData()->Cmd8, dat1, dat2,
                                dat3, dat4);
}

bool sendWRcmd_setSendBlock160(uint8_t leng, uint8_t const *buf) {
    SharedMemBuffer::getData()->next.sendAsyDatLen = std::min(leng, (uint8_t)160);
    clear(SharedMemBuffer::getData()->next.sendAsynchDataBuf);
    memcpy((char *)(&SharedMemBuffer::getData()->next.sendAsynchDataBuf),
           (char const *)(buf),
           SharedMemBuffer::getDataConst()->next.sendAsyDatLen);
    return true;                // ok, will be sent
}
uint8_t sendWRcmd_getSendBlock160(uint8_t *leng, uint8_t *buf) {
    //qDebug()  << "pi gpi: restoring send data";
    *leng = SharedMemBuffer::getDataConst()->next.sendAsyDatLen;
    memcpy((char *)buf,
           (char const *)(SharedMemBuffer::getData()->next.sendAsynchDataBuf),
           *leng);
    SharedMemBuffer::getData()->next.sendAsyDatLen = 0;
    return *leng;
}


// ------------------------------------------------------------------------------------
// MDB Sendind Data  are store here for next transport to DC (Device Controller)
// Transport to Slave runs every 100ms, answer from mdb-slave (e.g. coin changer) comes rigth
// with next slave answer

// start with: SENDDIRCMD_EXCHGMDB,
    // send crude data from here to DC, DC to mdb slaves, mdb answer, return here within 50ms

uint8_t epi_store64ByteSendData(uint8_t length, uint8_t const *buf) {
    // HWapi writes data to be forwarded to DC and further to mdb-device
    memcpy((char *)(&SharedMemBuffer::getData()->Sdata.mdbSendBuffer[0]),
           (char const *)buf, length);
    SharedMemBuffer::getData()->Sdata.mdbSendLen = length;
    return 0;
}

uint8_t gpi_restore64ByteSendData(uint8_t *length, uint8_t *buf) {
    // datif reads data to forward to dc
    memcpy((char *)(buf),
           (char const *)(&SharedMemBuffer::getDataConst()->Sdata.mdbSendBuffer[0]),
            SharedMemBuffer::getDataConst()->Sdata.mdbSendLen);
    *length = SharedMemBuffer::getDataConst()->Sdata.mdbSendLen;
    SharedMemBuffer::getData()->Sdata.mdbSendLen = 0;
    return 0;
}

//------------------------------------------------------------------------------------
//------------------------------------------------------------------------------------
//---------------------------------------- Printer Text Fifo -------------------------
void epi_storeUserOfSendingTextBuffer(uint8_t user, uint8_t para1,
                                      uint8_t para2, uint8_t para3,
                                      uint8_t para4 ) {
    // user=1: Text-Print is using this buffer
    //      2: QR-code-Printer is using this buffer
    SharedMemBuffer::getData()->prn.DataBufferUser = user;
    SharedMemBuffer::getData()->prn.DataParameters[0] = para1;
    SharedMemBuffer::getData()->prn.DataParameters[1] = para2;
    SharedMemBuffer::getData()->prn.DataParameters[2] = para3;
    SharedMemBuffer::getData()->prn.DataParameters[3] = para4;
}

uint8_t gpi_getUserOfSendingTextBuffer(uint8_t *para1, uint8_t *para2,
                                       uint8_t *para3, uint8_t *para4) {
    // user=1: Text-Print is using this buffer
    //      2: QR-code-Printer is using this buffer
    //qDebug() << "returning user "<< prnDataBufferUser;
    *para1 = SharedMemBuffer::getDataConst()->prn.DataParameters[0];
    *para2 = SharedMemBuffer::getDataConst()->prn.DataParameters[1];
    *para3 = SharedMemBuffer::getDataConst()->prn.DataParameters[2];
    *para4 = SharedMemBuffer::getDataConst()->prn.DataParameters[3];
    return SharedMemBuffer::getDataConst()->prn.DataBufferUser;
}

// Sending Text Fifo
// ONE printer doc consists of 20 x 64 byte
// #define MAXNROF_PRNBYTES   64
// #define MAXNROF_PRNBLOCKS  20
// static char Sdata_PRN_TEXT[MAXNROF_PRNBLOCKS][MAXNROF_PRNBYTES];
// static uint8_t pPrnDataBuff;        // points to next PRINTER_BLOCK

//static uint8_t pPrnDataBuff;        // points to next waiting printer text
//  defined above, needed if more then one text is stored (before sent)
// every block will be sent after 100ms, if 8 blocks are stored within this 100ms
// then pointer goes up to 8. Important: FIFO!!!!!!!!

void epi_resetPrinterStack(void) {
    SharedMemBuffer::getData()->prn.pDataBuff = 0;
}

uint8_t epi_storePrnText(char *buf, uint8_t leng) {
    // store text from Gui in next higher free memory 0....9

    uint16_t const len = std::min(leng, (uint8_t)MAXNROF_PRNBYTES);
    int const pp = SharedMemBuffer::getDataConst()->prn.pDataBuff;

    // next free memory block with 64byte each
    if (pp < MAXNROF_PRNBLOCKS) {
        clear(SharedMemBuffer::getData()->Sdata.PRN_TEXT[pp]);
        // copy new text into buffer
        memcpy((char *)(&SharedMemBuffer::getData()->Sdata.PRN_TEXT[pp][0]),
               (char const *)(buf), len);
        // inc pointer if end not yet reached
        SharedMemBuffer::getData()->prn.pDataBuff++;
        return 0;       // OK
    }

    return 1;   // not possible, no free mem
}

uint8_t gpi_restorePrnText(uint8_t *retbuf) {
    // read printer text and send to slave, size of retbuf == 64

    // always read from [0] because this is the oldest (Fifo)
    // then move all text lines down by one and dec pointer
    int pp = SharedMemBuffer::getDataConst()->prn.pDataBuff;

    if (pp > 0) { // next free memory block with 64byte each
        // example: pp=5: then buffers [0...4] are occupied

        memcpy((char *)retbuf, // restore oldest text
               (char const *)(&SharedMemBuffer::getData()->Sdata.PRN_TEXT[0][0]),
               sizeof(SharedMemBuffer::getDataConst()->Sdata.PRN_TEXT[0]));

        // now copy textline [1] to [0], then
        // copy textline [2] to [1], then
        // copy textline [3] to [2] .... upto [pp-1] to [pp-2]
        // hint: copying from 9....0 would delete all strings!!!!!!

        for (int nn=0; nn < (pp-1); nn++) {
            memmove((char *)(&SharedMemBuffer::getData()->Sdata.PRN_TEXT[nn+1][0]),
                    (char const *)(&SharedMemBuffer::getDataConst()->Sdata.PRN_TEXT[nn][0]),
                    sizeof(SharedMemBuffer::getDataConst()->Sdata.PRN_TEXT[0]));
        }

        if (SharedMemBuffer::getDataConst()->prn.pDataBuff > 0) {
            SharedMemBuffer::getData()->prn.pDataBuff--;
        }

        pp = SharedMemBuffer::getDataConst()->prn.pDataBuff;
        // example: pp=4: then buffers [0...3] are still occupied, pp=0: all buffers empty

        // now clear highest copyed line (which got free now)
        clear(SharedMemBuffer::getData()->Sdata.PRN_TEXT[pp]);

        // optionally: clear all remaining higher lines:
        for (int nn = (pp+1); nn < MAXNROF_PRNBLOCKS; nn++) {
            clear(SharedMemBuffer::getData()->Sdata.PRN_TEXT[nn]);
        }
        return 0;
    }

    return 1; // no text in buffer
}

uint8_t gpi_chk4remainingText(void) {
    // retval: 0: no more textline left (to send)  >0: nr of 64byte-blocks
    return SharedMemBuffer::getDataConst()->prn.pDataBuff;
}

// ---------------------------------------------------------------------------------
//   11.4.23 neu, Kommando direkt an "FastDevice"-protokoll senden, nicht mehr umsetzen
// ---------------------------------------------------------------------------------

/* convention: use simple (not rotating) FIFO Stack:
Example: nrOfCmdsInQueue=4 then
    nextAsynchsendCmd0[0]=cmd1      // was stored as first
    nextAsynchsendCmd0[1]=cmd2
    nextAsynchsendCmd0[2]=cmd3
    nextAsynchsendCmd0[3]=cmd4      // came in as last

    Send: [0] first, then move buffer 1 down:
    nextAsynchsendCmd0[0]=cmd2
    nextAsynchsendCmd0[1]=cmd3
    nextAsynchsendCmd0[2]=cmd4
    nextAsynchsendCmd0[3]=0;
    nrOfCmdsInQueue=3 now
*/

void sendFDcmd_clrStack(void) {
    clear(SharedMemBuffer::getData()->FDShort.wrCmd);
    clear(SharedMemBuffer::getData()->FDShort.rdCmd);
    clear(SharedMemBuffer::getData()->FDShort.blkNr);
    clear(SharedMemBuffer::getData()->FDShort.para1);
    clear(SharedMemBuffer::getData()->FDShort.para2);
    clear(SharedMemBuffer::getData()->FDShort.para3);
    clear(SharedMemBuffer::getData()->FDShort.para4);
    SharedMemBuffer::getData()->FDShort.cmdsInQueue = 0;
}

bool sendFDcmd_set(uint8_t nextWrCmd, uint8_t nextRdCmd, uint8_t blockNum,
                   uint8_t dat1, uint8_t dat2, uint8_t dat3, uint8_t dat4) {
    // write Command to memory, wait for transport
    int const c = SharedMemBuffer::getDataConst()->FDShort.cmdsInQueue;
    if (c < FDCMD_STACKDEPTH) {
        SharedMemBuffer::getData()->FDShort.wrCmd[c] = nextWrCmd;
        SharedMemBuffer::getData()->FDShort.rdCmd[c] = nextRdCmd;
        SharedMemBuffer::getData()->FDShort.blkNr[c] = blockNum;
        SharedMemBuffer::getData()->FDShort.para1[c] = dat1;
        SharedMemBuffer::getData()->FDShort.para2[c] = dat2;
        SharedMemBuffer::getData()->FDShort.para3[c] = dat3;
        SharedMemBuffer::getData()->FDShort.para4[c] = dat4;
        SharedMemBuffer::getData()->FDShort.cmdsInQueue = (c + 1);
        return true;                // ok, will be sent
    }
    qCritical() << "cannot save cmd because stack is full";
    return false; // not possible
}

bool sendFDcmd_get(uint8_t *nextWrCmd, uint8_t *nextRdCmd, uint8_t *blockNum,
                   uint8_t *dat1, uint8_t *dat2, uint8_t *dat3, uint8_t *dat4) {

    int const c = SharedMemBuffer::getDataConst()->FDShort.cmdsInQueue;

    if ((c > 0)  && (c <= FDCMD_STACKDEPTH)) {
        *nextWrCmd = SharedMemBuffer::getDataConst()->FDShort.wrCmd[c];
        *nextRdCmd = SharedMemBuffer::getDataConst()->FDShort.rdCmd[c];
        *blockNum = SharedMemBuffer::getDataConst()->FDShort.blkNr[c];
        *dat1 = SharedMemBuffer::getDataConst()->FDShort.para1[c];
        *dat2 = SharedMemBuffer::getDataConst()->FDShort.para2[c];
        *dat3 = SharedMemBuffer::getDataConst()->FDShort.para3[c];
        *dat4 = SharedMemBuffer::getDataConst()->FDShort.para4[c];

        // move Puffer down by one element
        for (int m = 0; m < c-1; ++m) {
            int const n = m + 1;
            SharedMemBuffer::getData()->FDShort.wrCmd[m] = SharedMemBuffer::getDataConst()->FDShort.wrCmd[n];
            SharedMemBuffer::getData()->FDShort.rdCmd[m] = SharedMemBuffer::getDataConst()->FDShort.rdCmd[n];
            SharedMemBuffer::getData()->FDShort.blkNr[m] = SharedMemBuffer::getDataConst()->FDShort.blkNr[n];
            SharedMemBuffer::getData()->FDShort.para1[m] = SharedMemBuffer::getDataConst()->FDShort.para1[n];
            SharedMemBuffer::getData()->FDShort.para2[m] = SharedMemBuffer::getDataConst()->FDShort.para2[n];
            SharedMemBuffer::getData()->FDShort.para3[m] = SharedMemBuffer::getDataConst()->FDShort.para3[n];
            SharedMemBuffer::getData()->FDShort.para4[m] = SharedMemBuffer::getDataConst()->FDShort.para4[n];
        }
        SharedMemBuffer::getData()->FDShort.cmdsInQueue = c - 1;
        return true;                // ok, will be sent
    }
    return false;           // not possible
}

uint8_t check4FDshortCmd(void) { // returns number of waiting command, max FDCMD_STACKDEPTH
    return SharedMemBuffer::getDataConst()->FDShort.cmdsInQueue;
}


uint8_t check4freeFDshortCmd(void) { // returns number of free places in short-command stack
    return FDCMD_STACKDEPTH - SharedMemBuffer::getDataConst()->FDShort.cmdsInQueue;
}

void longFDcmd_clrStack(void) {
    clear(SharedMemBuffer::getData()->FDLong.wrCmd);
    clear(SharedMemBuffer::getData()->FDLong.rdCmd);
    clear(SharedMemBuffer::getData()->FDLong.blkNr);
    clear(SharedMemBuffer::getData()->FDLong.length);
    clear(SharedMemBuffer::getData()->FDLong.para);
    SharedMemBuffer::getData()->FDLong.cmdsInQueue = 0;
}

bool longFDcmd_set(uint8_t nextWrCmd, uint8_t nextRdCmd, uint8_t blockNum,
                   uint8_t length, uint8_t const *data) {
    // write Command to memory, wait for transport
    // data buffer size always 64! data[64], padded with 0
    int const c = SharedMemBuffer::getDataConst()->FDLong.cmdsInQueue;

    if (c > 0 && c <= FDLONG_STACKDEPTH) {
        SharedMemBuffer::getData()->FDLong.wrCmd[c] = nextWrCmd;
        SharedMemBuffer::getData()->FDLong.rdCmd[c] = nextRdCmd;
        SharedMemBuffer::getData()->FDLong.blkNr[c] = blockNum;
        SharedMemBuffer::getData()->FDLong.length[c] = length;
        memcpy((char *)SharedMemBuffer::getData()->FDLong.para[c],
               (char const *)data,
                sizeof(SharedMemBuffer::getDataConst()->FDLong.para[0]));

        SharedMemBuffer::getData()->FDLong.cmdsInQueue = c + 1;
        return true;                // ok, will be sent
    }
    qCritical() << "cannot save cmd because stack is full";
    return false;           // not possible
}

bool longFDcmd_get(uint8_t *nextWrCmd, uint8_t *nextRdCmd, uint8_t *blockNum,
                   uint8_t *length, uint8_t *data) {

    int const c = SharedMemBuffer::getDataConst()->FDLong.cmdsInQueue;

    if (c > 0 && c <= FDLONG_STACKDEPTH) {
        *nextWrCmd = SharedMemBuffer::getDataConst()->FDLong.wrCmd[c];
        *nextRdCmd = SharedMemBuffer::getDataConst()->FDLong.rdCmd[c];
        *blockNum  = SharedMemBuffer::getDataConst()->FDLong.blkNr[c];
        *length    = SharedMemBuffer::getDataConst()->FDLong.length[c];
        memcpy((char *)data,
               (char const *)SharedMemBuffer::getDataConst()->FDLong.para[c],
                sizeof(SharedMemBuffer::getDataConst()->FDLong.para[0]));

        // move Puffer down by one element
        for (int m = 0; m < c-1; ++m) {
            int const n = m + 1;
            SharedMemBuffer::getData()->FDLong.wrCmd[m] = SharedMemBuffer::getDataConst()->FDLong.wrCmd[n];
            SharedMemBuffer::getData()->FDLong.rdCmd[m] = SharedMemBuffer::getDataConst()->FDLong.rdCmd[n];
            SharedMemBuffer::getData()->FDLong.blkNr[m] = SharedMemBuffer::getDataConst()->FDLong.blkNr[n];
            SharedMemBuffer::getData()->FDLong.length[m] = SharedMemBuffer::getData()->FDLong.length[n];
            memcpy((char *)SharedMemBuffer::getData()->FDLong.para[m],
                   (char const *)SharedMemBuffer::getData()->FDLong.para[n],
                   sizeof(SharedMemBuffer::getDataConst()->FDLong.para[0]));
        }
        SharedMemBuffer::getData()->FDLong.cmdsInQueue = c - 1;
        return true;                // ok, will be sent
    }

    qCritical() << "cannot save cmd because stack is full";
    return false;           // not possible
}

uint8_t check4FDlongCmd(void) { // returns number of waiting command
    return SharedMemBuffer::getDataConst()->FDLong.cmdsInQueue;
}

uint8_t check4freeFDlongCmd(void) { // returns number of free places in command stack
    return FDLONG_STACKDEPTH - SharedMemBuffer::getDataConst()->FDLong.cmdsInQueue;
}
