#include "shared_mem_buffer.h"

#include <QDebug>
#include <QCoreApplication>
#include <atomic>

#ifdef QT_POSIX_IPC
    // The POSIX backend can be explicitly selected using the -feature-ipc_posix
    // option to the Qt configure script. If it is enabled, the QT_POSIX_IPC
    // macro will be defined. -> we use SystemV shared memory
#error "QT_POSIX_IPC defined"
#else
#include <sys/ipc.h>    // ftok
#endif

#ifdef __linux__
#include <unistd.h>
#endif

#define PROCESS_SHARED (PTHREAD_PROCESS_SHARED)

// std::atomic_bool SharedMemBuffer::__sharedMemLocked{false};

QSharedMemory *SharedMemBuffer::getShm(std::size_t size) {
    static QSharedMemory shMem;
    if (size > 0) {
        static const long nativeKey = ftok("/etc/os-release", 'H');
        static const QString fkey = std::to_string(nativeKey).c_str();
        shMem.setKey(fkey);
        if (!shMem.isAttached()) {
            if (shMem.create(size)) {
                // the creator is the owner of the shared-memory segment
                SharedMemBuffer *shm = (SharedMemBuffer *)(shMem.data());
                if (shm) {
                    shm->init = 0;
                    // mark the shm
                    memset(shm->appName, 0x00, sizeof(shm->appName));
                    strncpy(shm->appName,
                            QCoreApplication::applicationName().toLocal8Bit().constData(),
                            sizeof(shm->appName)-1);
                    shm->pid = getpid();

                    // create condition variables plus associated mutexes for
                    // sending/receiving commands to dc
                    int r = 0;
                    if ((r = pthread_mutexattr_init(&shm->sndMtxA)) != 0) {
                        qCritical() << "init mutex attribute"
                                    << strerror(errno) << r;
                    }
                    if ((r == 0) &&
                        (r = pthread_mutexattr_setpshared(
                                &shm->sndMtxA, PTHREAD_PROCESS_SHARED) != 0)) {
                        qCritical() << "set s-mutex attribute to process shared"
                                    << strerror(errno) << r;
                    }
                    if ((r == 0) &&
                        (r = pthread_mutex_init(&shm->sndMtx,
                                                &shm->sndMtxA) != 0)) {
                        qCritical() << "init s-mutex" << strerror(errno) << r;
                    }
                    if ((r == 0) &&
                        (r = pthread_condattr_init(&shm->sndCndA) != 0)) {
                        qCritical() << "init s-mutex" << strerror(errno) << r;
                    }
                    if ((r == 0) &&
                        (r = pthread_condattr_setpshared(
                                &shm->sndCndA, PTHREAD_PROCESS_SHARED) != 0)) {
                        qCritical() << "init s-cond attribute to process shared"
                                    << strerror(errno) << r;
                    }
                    if ((r == 0) &&
                        (r = pthread_cond_init(&shm->sndCndV,
                                               &shm->sndCndA) != 0)) {
                        qCritical() << "init s-condvar" << strerror(errno) << r;
                    }

                    if ((r = pthread_mutexattr_init(&shm->rcvMtxA)) != 0) {
                        qCritical() << "init r-mutex attribute"
                                    << strerror(errno) << r;
                    }
                    if ((r == 0) &&
                        (r = pthread_mutexattr_setpshared(
                                &shm->rcvMtxA, PTHREAD_PROCESS_SHARED) != 0)) {
                        qCritical() << "set r-mutex attribute to process shared"
                                    << strerror(errno) << r;
                    }
                    if ((r == 0) &&
                        (r = pthread_mutex_init(&shm->rcvMtx,
                                                &shm->rcvMtxA) != 0)) {
                        qCritical() << "init r-mutex" << strerror(errno) << r;
                    }
                    if ((r == 0) &&
                        (r = pthread_condattr_init(&shm->rcvCndA) != 0)) {
                        qCritical() << "init r-mutex" << strerror(errno) << r;
                    }
                    if ((r == 0) &&
                        (r = pthread_condattr_setpshared(
                                &shm->rcvCndA, PTHREAD_PROCESS_SHARED) != 0)) {
                        qCritical() << "init r-cond attribute to process shared"
                                    << strerror(errno) << r;
                    }
                    if ((r == 0) &&
                        (r = pthread_cond_init(&shm->rcvCndV,
                                               &shm->rcvCndA) != 0)) {
                        qCritical() << "init s-condvar" << strerror(errno) << r;
                    }
                    shm->init = 1;
                }
                return &shMem;
            } else {
                if (shMem.error() == QSharedMemory::AlreadyExists) {
                    if (shMem.attach()) {
                        return &shMem;
                    }
                }
            }
            qCritical() << shMem.nativeKey() << shMem.key() << shMem.data()
                        << shMem.error() << shMem.errorString();
            return nullptr;
        }
    }
    return &shMem;
}
